import 'dart:isolate';
import 'dart:ui';
import 'package:flutter/foundation.dart';
import 'package:flutter_downloader/flutter_downloader.dart';
import 'ALDownloaderHandlerInterface.dart';
import 'ALDownloaderPersistentFileManager.dart';
import 'ALDownloaderStatus.dart';
import 'dart:io';

/// ALDownloader
class ALDownloader {
  /// initialize
  ///
  /// call or lazy call
  ///
  /// lazy call: when call [download]
  static Future<void> initialize() async {
    if (!_isInitial) {
      // FlutterDownloader initialize
      await FlutterDownloader.initialize(debug: false, ignoreSsl: true);

      // register FlutterDownloader callback
      FlutterDownloader.registerCallback(_downloadCallback);

      // register the isolate communication service
      _addIsolateNameServerPortService();

      // extract all current tasks from database and execute the tasks that need to be executed
      await _loadAndTryToRunTask();

      // a dirty flag that guarantees that this scope is executed only once
      _isInitial = true;
    }
  }

  /// download
  ///
  /// **parameters**
  ///
  /// [url] url
  ///
  /// [downloaderHandlerInterface] downloader handler interface
  static Future<void> download(String? url,
      {ALDownloaderHandlerInterface? downloaderHandlerInterface}) async {
    if (url == null) throw "ALDownloader | download error = url is null";

    if (downloaderHandlerInterface != null) {
      final aBinder =
          _ALDownloaderBinder(url, downloaderHandlerInterface, false);
      _alDownloaderBinders.add(aBinder);
    }

    await initialize();

    var anALDownloadTask = _getALDownloadTaskFromUrl(url);

    if (anALDownloadTask != null) {
      if (await _isShouldRemoveDataForSavedDir(
          anALDownloadTask.savedDir, url, anALDownloadTask.status)) {
        await _innerRemove(url);
        anALDownloadTask = null;
      }
    }

    if (anALDownloadTask == null) {
      // get the 'physical directory path' and 'file name' of the file by URL
      final alDownloaderPathComponentModel =
          await ALDownloaderPersistentFileManager
              .lazyGetALDownloaderPathModelForUrl(url);

      final dir = alDownloaderPathComponentModel.dir;

      // equeued a task
      final taskId = await FlutterDownloader.enqueue(
          url: url,
          savedDir: dir,
          fileName: alDownloaderPathComponentModel.fileName,
          showNotification: false,
          openFileFromNotification: false);

      if (taskId != null)
        _addALDownloadTaskOrReplaceALDownloadTaskId(
            url, taskId, DownloadTaskStatus.enqueued, 0, dir);

      debugPrint(
          "ALDownloader | a download task was generated, download status = enqueued, taskId = $taskId");
    } else if (anALDownloadTask.status == DownloadTaskStatus.failed ||
        anALDownloadTask.status == DownloadTaskStatus.canceled) {
      final newTaskIdForRetry =
          await FlutterDownloader.retry(taskId: anALDownloadTask.taskId);
      if (newTaskIdForRetry != null) {
        debugPrint("ALDownloader | newTaskIdForRetry = $newTaskIdForRetry");
        _addALDownloadTaskOrReplaceALDownloadTaskId(url, newTaskIdForRetry,
            DownloadTaskStatus.undefined, anALDownloadTask.progress, "");
      }
    } else if (anALDownloadTask.status == DownloadTaskStatus.paused) {
      final newTaskIdForResume =
          await FlutterDownloader.resume(taskId: anALDownloadTask.taskId);
      if (newTaskIdForResume != null)
        _addALDownloadTaskOrReplaceALDownloadTaskId(url, newTaskIdForResume,
            DownloadTaskStatus.undefined, anALDownloadTask.progress, "");
    } else if (anALDownloadTask.status == DownloadTaskStatus.complete) {
      debugPrint(
          "ALDownloader | try to download downloadtaskOfUrl = $url, url is complete");
      _alDownloaderBinders.forEach((element) {
        if (element.forUrl == url) {
          final progressHandler =
              element.downloaderHandlerHolder.progressHandler;
          if (progressHandler != null) {
            // ignore: non_constant_identifier_names
            final int_progress =
                anALDownloadTask == null ? 0 : anALDownloadTask.progress;
            // ignore: non_constant_identifier_names
            final double_progress =
                double.tryParse(((int_progress / 100).toStringAsFixed(2))) ?? 0;
            progressHandler(double_progress);
          }

          final succeededHandler =
              element.downloaderHandlerHolder.succeededHandler;
          if (succeededHandler != null) succeededHandler();
        }
      });
      _alDownloaderBinders.removeWhere(
          (element) => element.forUrl == url && !element.isForever);
    } else if (anALDownloadTask.status == DownloadTaskStatus.running) {
      debugPrint(
          "ALDownloader | try to download downloadtaskOfUrl = $url, but url is running");
    } else if (anALDownloadTask.status == DownloadTaskStatus.enqueued) {
      debugPrint(
          "ALDownloader | try to download downloadtaskOfUrl = $url, but url is enqueued");
    } else {
      debugPrint(
          "ALDownloader | try to download downloadtaskOfUrl = $url, but url is undefined");
    }
  }

  /// add a downloader handler interface
  ///
  /// **parameters**
  ///
  /// [downloaderHandlerInterface] downloader handler interface
  ///
  /// [forUrl] url
  static void addDownloaderHandlerInterface(
      ALDownloaderHandlerInterface? downloaderHandlerInterface,
      String? forUrl) {
    if (downloaderHandlerInterface == null || forUrl == null) return;
    final aBinder =
        _ALDownloaderBinder(forUrl, downloaderHandlerInterface, false);
    _alDownloaderBinders.add(aBinder);
  }

  /// add a forever downloader handler interface
  ///
  /// **parameters**
  ///
  /// [downloaderHandlerInterface] downloader handler interface
  ///
  /// [forUrl] url
  static void addForeverDownloaderHandlerInterface(
      ALDownloaderHandlerInterface? downloaderHandlerInterface,
      String? forUrl) {
    if (downloaderHandlerInterface == null || forUrl == null) return;
    final aBinder =
        _ALDownloaderBinder(forUrl, downloaderHandlerInterface, true);
    _alDownloaderBinders.add(aBinder);
  }

  /// remove a downloader handler interface
  ///
  /// **parameters**
  ///
  /// [url] url
  static void removeDownloaderHandlerInterfaceForUrl(String url) =>
      _alDownloaderBinders.removeWhere((element) => url == element.forUrl);

  /// remove all downloader handler interfaces
  static void removeDownloaderHandlerInterfaceForAll() =>
      _alDownloaderBinders.clear();

  /// get download status for [url]
  ///
  /// **parameters**
  ///
  /// [url] url
  ///
  /// **return**
  ///
  /// download status [ALDownloaderStatus]
  static ALDownloaderStatus getDownloadStatusForUrl(String url) {
    ALDownloaderStatus alDownloaderStatus;

    try {
      final anALDownloadTask = _getALDownloadTaskFromUrl(url);
      final downloadTaskStatus = anALDownloadTask?.status;

      if (downloadTaskStatus == null ||
          downloadTaskStatus == DownloadTaskStatus.canceled ||
          downloadTaskStatus == DownloadTaskStatus.undefined ||
          downloadTaskStatus == DownloadTaskStatus.enqueued)
        alDownloaderStatus = ALDownloaderStatus.unstarted;
      else if (downloadTaskStatus == DownloadTaskStatus.running)
        alDownloaderStatus = ALDownloaderStatus.downloading;
      else if (downloadTaskStatus == DownloadTaskStatus.paused)
        alDownloaderStatus = ALDownloaderStatus.paused;
      else if (downloadTaskStatus == DownloadTaskStatus.failed)
        alDownloaderStatus = ALDownloaderStatus.failed;
      else
        alDownloaderStatus = ALDownloaderStatus.succeeded;
    } catch (error) {
      alDownloaderStatus = ALDownloaderStatus.unstarted;
      debugPrint(
          "ALDownloader | getDownloadStatusForUrl = $url, error = $error");
    }
    return alDownloaderStatus;
  }

  /// get download progress for [url]
  ///
  /// **parameters**
  ///
  /// [url] url
  ///
  /// **return**
  ///
  /// download progress
  static double getDownloadProgressForUrl(String url) {
    double alDownloaderProgress;

    try {
      final anALDownloadTask = _getALDownloadTaskFromUrl(url);

      int progress = (anALDownloadTask == null ? 0 : anALDownloadTask.progress);

      // ignore: non_constant_identifier_names
      alDownloaderProgress =
          double.tryParse(((progress / 100).toStringAsFixed(2))) ?? 0;
    } catch (error) {
      alDownloaderProgress = 0;
      debugPrint(
          "ALDownloader | get download progress for url = $url, error = $error");
    }
    return alDownloaderProgress;
  }

  /// get completed key value pairs
  ///
  /// urls for resources that download succeeded or failed are added to the pool
  ///
  /// **return**
  ///
  /// [url: bool] - completed key value pairs
  static Map<String, bool> get completedKVs {
    final Map<String, bool> aMap = {};
    try {
      _alDownloadTasks.forEach((element) {
        if (element.status == DownloadTaskStatus.complete) {
          aMap[element.url] = true;
        } else if (element.status == DownloadTaskStatus.failed) {
          aMap[element.url] = false;
        }
      });
    } catch (error) {
      debugPrint("ALDownloader | get completedKVs error = $error");
    }
    return aMap;
  }

  /// cancel download
  ///
  /// the running download and incomplete bytes will be deleted
  ///
  /// **parameters**
  ///
  /// [url] url
  static Future<void> cancel(String url) async {
    assert(_isInitial,
        "ALDownloader | ALDownloader.initialize or ALDownloader.download must be called first");

    final alDownloadTask = _getALDownloadTaskFromUrl(url);

    if (alDownloadTask == null) return;

    if (alDownloadTask.status == DownloadTaskStatus.running)
      await _removeCommonTask(alDownloadTask);
  }

  /// cancel all downloads
  ///
  /// this is a multiple of [cancel], see [cancel]
  static Future<void> cancelAll() async {
    final aTemp = [];
    aTemp.addAll(_alDownloadTasks);
    for (final alDownloadTask in aTemp) {
      final url = alDownloadTask.url;
      await cancel(url);
    }
  }

  /// pause download
  ///
  /// the incomplete bytes will not be deleted
  ///
  /// undefined/enqueued task will be removed from the download queue
  ///
  /// **parameters**
  ///
  /// [url] url
  static Future<void> pause(String url) async {
    assert(_isInitial,
        "ALDownloader | ALDownloader.initialize or ALDownloader.download must be called first");

    try {
      final alDownloadTask = _getALDownloadTaskFromUrl(url);

      if (alDownloadTask == null) {
        debugPrint(
            "ALDownloader | pause, url = $url, but url's alDownloadTask is null");
        return;
      }

      debugPrint(
          "ALDownloader | pause, url = $url, alDownloadTask.status = ${alDownloadTask.status}");

      final taskId = alDownloadTask.taskId;
      if (alDownloadTask.status == DownloadTaskStatus.undefined ||
          alDownloadTask.status == DownloadTaskStatus.enqueued) {
        _removeUndefinedEnqueuedTask(alDownloadTask);
      } else if (alDownloadTask.status == DownloadTaskStatus.running) {
        await FlutterDownloader.pause(taskId: taskId);
      }
    } catch (error) {
      debugPrint("ALDownloader | pause error = $error");
    }
  }

  /// pause all downloads
  ///
  /// this is a multiple of [pause], see [pause]
  static Future<void> pauseAll() async {
    final aTemp = [];
    aTemp.addAll(_alDownloadTasks);
    for (final alDownloadTask in aTemp) {
      final url = alDownloadTask.url;
      await pause(url);
    }
  }

  /// remove data
  ///
  /// it will automatically cancel the downloading url
  ///
  /// **parameters**
  ///
  /// [url] url
  static Future<void> remove(String url) async {
    assert(_isInitial,
        "ALDownloader | ALDownloader.initialize or ALDownloader.download must be called first");

    await _innerRemove(url);
  }

  /// remove all data
  ///
  /// this is a multiple of [remove], see [remove]
  static Future<void> removeAll() async {
    final aTemp = [];
    aTemp.addAll(_alDownloadTasks);
    for (final alDownloadTask in aTemp) {
      final url = alDownloadTask.url;
      await remove(url);
    }
  }

  /// ---------------------- Private API ----------------------

  /// maintain custom download tasks
  ///
  /// purpose: avoid frequent I/O
  ///
  /// add a new custom task or update the [url] mapping of the existing task's taskId
  static void _addALDownloadTaskOrReplaceALDownloadTaskId(String? url,
      String taskId, DownloadTaskStatus status, int progress, String savedDir) {
    if (url == null) {
      debugPrint(
          "_addALDownloadTaskOrReplaceALDownloadTaskId error url == null");
      return;
    }

    _ALDownloadTask? anALDownloadTask;

    try {
      anALDownloadTask =
          _alDownloadTasks.firstWhere((element) => element.url == url);
      if (savedDir != "") anALDownloadTask.savedDir = savedDir;
      anALDownloadTask.taskId = taskId;
      anALDownloadTask.status = status;
      anALDownloadTask.progress = progress;
    } catch (error) {
      debugPrint(
          "ALDownloader | _addALDownloadTaskOrReplaceALDownloadTaskId error = $error");
    }

    if (anALDownloadTask == null) {
      anALDownloadTask = _ALDownloadTask(url);
      if (savedDir != "") anALDownloadTask.savedDir = savedDir;
      anALDownloadTask.taskId = taskId;
      anALDownloadTask.status = status;
      anALDownloadTask.progress = progress;

      _alDownloadTasks.add(anALDownloadTask);
    }
  }

  /// register send port and receive port for [IsolateNameServer]
  ///
  /// used for communication between download isolate and the main isolate
  static void _addIsolateNameServerPortService() {
    IsolateNameServer.registerPortWithName(
        _receivePort.sendPort, _kDownloaderSendPort);
    _receivePort.listen((dynamic data) {
      final id = data[0];
      final status = data[1];
      final progress = data[2];
      _processDataFromPort(id, status, progress);
    });
  }

  /// the callback binded by [FlutterDownloader]
  static void _downloadCallback(
      String id, DownloadTaskStatus status, int progress) {
    final SendPort? send =
        IsolateNameServer.lookupPortByName(_kDownloaderSendPort);
    send?.send([id, status, progress]);
  }

  /// process the [FlutterDownloader]'s callback
  static void _processDataFromPort(
      String id, DownloadTaskStatus status, int progress) {
    final alDownloadTask = _getALDownloadTaskFromTaskId(id);

    if (alDownloadTask == null) {
      debugPrint("not found alDownloadTask, id = $id");
      return;
    }

    final url = alDownloadTask.url;

    _addALDownloadTaskOrReplaceALDownloadTaskId(url, id, status, progress, "");

    // ignore: non_constant_identifier_names
    final double_progress =
        double.tryParse(((progress / 100).toStringAsFixed(2))) ?? 0;

    if (status == DownloadTaskStatus.complete) {
      debugPrint(
          "ALDownloader | _downloadCallback download succeeded, url = $url, id = $id");

      _alDownloaderBinders.forEach((element) {
        if (element.forUrl == url) {
          final progressHandler =
              element.downloaderHandlerHolder.progressHandler;
          if (progressHandler != null) progressHandler(double_progress);

          final succeededHandler =
              element.downloaderHandlerHolder.succeededHandler;
          if (succeededHandler != null) succeededHandler();
        }
      });
    } else if (status == DownloadTaskStatus.running) {
      _alDownloaderBinders.forEach((element) {
        if (element.forUrl == url) {
          final progressHandler =
              element.downloaderHandlerHolder.progressHandler;
          if (progressHandler != null) progressHandler(double_progress);
        }
      });
    } else if (status == DownloadTaskStatus.paused) {
      _alDownloaderBinders.forEach((element) {
        if (element.forUrl == url) {
          final pausedHandler = element.downloaderHandlerHolder.pausedHandler;
          if (pausedHandler != null) pausedHandler();
        }
      });
    } else if (status == DownloadTaskStatus.canceled ||
        status == DownloadTaskStatus.failed) {
      debugPrint(
          "ALDownloader | _downloadCallback download canceled/failed, status = $status, url = $url, id = $id");
      _alDownloaderBinders.forEach((element) {
        if (element.forUrl == url) {
          final progressHandler =
              element.downloaderHandlerHolder.progressHandler;
          if (progressHandler != null) progressHandler(double_progress);

          final failedHandler = element.downloaderHandlerHolder.failedHandler;
          if (failedHandler != null) failedHandler();
        }
      });

      if (status == DownloadTaskStatus.canceled)
        _alDownloadTasks.remove(alDownloadTask);
    }

    if (status == DownloadTaskStatus.complete ||
        status == DownloadTaskStatus.failed ||
        status == DownloadTaskStatus.canceled) {
      _alDownloaderBinders.removeWhere(
          (element) => element.forUrl == url && !element.isForever);
    }

    debugPrint(
        "ALDownloader | final | _downloadCallback, id = $id, url = $url, status = $status, progress = $progress, double_progress = $double_progress");
  }

  /// load [FlutterDownloader]'s local database task to the memory cache, and attempt to execute the tasks
  static Future<void> _loadAndTryToRunTask() async {
    final tasks = await FlutterDownloader.loadTasks();
    if (tasks != null) {
      debugPrint(
          "ALDownloader | _loadAndTryToRunTask `Flutterdownloader` tasks length = ${tasks.length}");

      tasks.forEach((element) {
        debugPrint(
            "ALDownloader | _loadAndTryToRunTask `Flutterdownloader` task element url = ${element.url}, status = ${element.status}, element url = ${element.taskId}");
      });

      for (final element in tasks) {
        final taskId = element.taskId;
        final taskUrl = element.url;
        final status = element.status;
        final savedDir = element.savedDir;
        final progress = element.progress;

        final anALDownloadTask = _ALDownloadTask(taskUrl);
        anALDownloadTask.savedDir = savedDir;
        anALDownloadTask.taskId = taskId;
        anALDownloadTask.status = status;
        anALDownloadTask.progress = progress;
        _alDownloadTasks.add(anALDownloadTask);

        if (await _isShouldRemoveDataForSavedDir(savedDir, taskUrl, status))
          await _removeCommonTask(anALDownloadTask);
      }
    }

    debugPrint(
        "ALDownloader | _loadAndTryToRunTask `ALDownloader` tasks length = ${_alDownloadTasks.length}");

    _alDownloadTasks.forEach((element) {
      debugPrint(
          "ALDownloader | _loadAndTryToRunTask `ALDownloader` task element url = ${element.url}, status = ${element.status}, element url = ${element.taskId}");
    });
  }

  /// verify the savedDir to determine whether to delete data
  static Future<bool> _isShouldRemoveDataForSavedDir(
      String savedDir, String url, DownloadTaskStatus status) async {
    if (!(await ALDownloader._isInRootPathForPath(savedDir))) return true;

    if (Platform.isAndroid) {
      final aBool = await ALDownloaderPersistentFileManager
          .isExistAbsolutePhysicalPathOfFileForUrl(url);
      return !aBool;
    } else if (Platform.isIOS) {
      if (status == DownloadTaskStatus.complete) {
        final aBool = await ALDownloaderPersistentFileManager
            .isExistAbsolutePhysicalPathOfFileForUrl(url);
        return !aBool;
      } else {
        return false;
      }
    } else {
      return true;
    }
  }

  /// whether path is in root path
  static Future<bool> _isInRootPathForPath(String path) async {
    if (path == "") return false;

    // for deleting old version's data
    if (path.contains("/flutter/al_")) return false;

    final isSavedDirInRootPath =
        await ALDownloaderPersistentFileManager.isInRootPathWithPath(path);

    return isSavedDirInRootPath;
  }

  /// get task from custom download tasks by [url]
  static _ALDownloadTask? _getALDownloadTaskFromUrl(String url) {
    _ALDownloadTask? anALDownloadTask;
    try {
      anALDownloadTask =
          _alDownloadTasks.firstWhere((element) => url == element.url);
    } catch (error) {
      debugPrint("ALDownloader | _getALDownloadTaskFromUrl, error = $error");
    }
    return anALDownloadTask;
  }

  /// get task id from custom download tasks by [url]
  // ignore: unused_element
  static String? _getTaskIdFromUrl(String url) {
    String? taskId;
    try {
      taskId =
          _alDownloadTasks.firstWhere((element) => url == element.url).taskId;
    } catch (error) {
      debugPrint("ALDownloader | _getTaskIdWith, error = $error");
    }
    return taskId;
  }

  /// get task from custom download tasks by [taskId]
  // ignore: unused_element
  static _ALDownloadTask? _getALDownloadTaskFromTaskId(String taskId) {
    _ALDownloadTask? anALDownloadTask;
    try {
      anALDownloadTask =
          _alDownloadTasks.firstWhere((element) => taskId == element.taskId);
    } catch (error) {
      debugPrint("ALDownloader | _getALDownloadTaskFromTaskId, error = $error");
    }
    return anALDownloadTask;
  }

  /// get url from custom download tasks by [taskId]
  // ignore: unused_element
  static String? _getUrlWithTaskId(String taskId) {
    String? url;
    try {
      url = _alDownloadTasks
          .firstWhere((element) => taskId == element.taskId)
          .url;
    } catch (error) {
      debugPrint("ALDownloader | _getUrlWithTaskId, error = $error");
    }
    return url;
  }

  static Future<void> _innerRemove(String url) async {
    try {
      final alDownloaderTask = _getALDownloadTaskFromUrl(url);

      if (alDownloaderTask == null) {
        debugPrint(
            "ALDownloader | _innerRemove, url = $url, url's _ALDownloadTask is null");
        return;
      }

      await _removeCommonTask(alDownloaderTask);
    } catch (error) {
      debugPrint("ALDownloader | _innerRemove, url = $url, error = $error");
    }
  }

  /// remove non undefined/enqueued task
  /// running task will call back in [_processDataFromPort], non running task will be deleted directly
  static Future<void> _removeCommonTask(_ALDownloadTask task) async {
    try {
      final status = task.status;
      final id = task.taskId;
      final url = task.url;

      if (status != DownloadTaskStatus.running) {
        _alDownloaderBinders.forEach((element) {
          if (element.forUrl == url) {
            final progressHandler =
                element.downloaderHandlerHolder.progressHandler;
            if (progressHandler != null) progressHandler(-0.01);

            final failedHandler = element.downloaderHandlerHolder.failedHandler;
            if (failedHandler != null) failedHandler();
          }
        });

        _alDownloaderBinders.removeWhere(
            (element) => element.forUrl == url && !element.isForever);

        _alDownloadTasks.remove(task);
      }

      await FlutterDownloader.remove(taskId: id, shouldDeleteContent: true);
    } catch (error) {
      debugPrint(
          "ALDownloader | _removeCommonTask, task = $task, error = $error");
    }
  }

  /// remove undefined/enqueued task
  /// no call back,
  static Future<void> _removeUndefinedEnqueuedTask(_ALDownloadTask task) async {
    try {
      _alDownloadTasks.remove(task);
      await FlutterDownloader.remove(
          taskId: task.taskId, shouldDeleteContent: true);
    } catch (error) {
      debugPrint(
          "ALDownloader | _removeUndefinedEnqueuedTask, task = $task, error = $error");
    }
  }

  /// a dirty flag that guarantees that this scope is executed only once
  static bool _isInitial = false;

  /// custom download tasks
  static final List<_ALDownloadTask> _alDownloadTasks = [];

  /// a binder list for binding element of url and download ininterface
  static final List<_ALDownloaderBinder> _alDownloaderBinders = [];

  /// send port key
  static final _kDownloaderSendPort = "al_downloader_send_port";

  /// receive port
  static final ReceivePort _receivePort = ReceivePort();

  /// privatize constructor
  ALDownloader._();
}

/// class of custom download task
class _ALDownloadTask {
  final String url;
  String savedDir = "";

  String taskId = "";
  int progress = 0;
  DownloadTaskStatus status = DownloadTaskStatus.undefined;

  _ALDownloadTask(this.url);
}

/// a binder for binding element of url and download ininterface
/// it may bind more elements in the future
class _ALDownloaderBinder {
  _ALDownloaderBinder(
      this.forUrl, this.downloaderHandlerHolder, this.isForever);
  final String forUrl;
  final ALDownloaderHandlerInterface downloaderHandlerHolder;
  final bool isForever;
}
